import itertools
import os
import time
from collections import defaultdict, Callable
import numpy as np
import matplotlib.pyplot as plt
from multiprocessing import Pool, cpu_count

from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister, Aer, assemble, BasicAer, transpile
from qiskit.aqua import QuantumInstance, aqua_globals
from qiskit.aqua.operators import PauliExpectation, CircuitSampler, StateFn, CircuitOp, CircuitStateFn, \
    MatrixExpectation, X, Y, Z, I

from Qcover.utils import get_graph_weights
from Qcover.backends import Backend
import warnings
warnings.filterwarnings("ignore")


class CircuitByQiskit(Backend):
    """generate a instance of CircuitByQiskit"""
    def __init__(self,
                 expectation_calc_method: str = "statevector", # or sample
                 research: str = "QAOA",
                 is_parallel: bool = None) -> None:
        """initialize a instance of CircuitByQiskit"""
        super(CircuitByQiskit, self).__init__()
        self._p = None
        self._origin_graph = None
        self._is_parallel = False if is_parallel is None else is_parallel
        self._expectation_calc_method = expectation_calc_method
        self._research = research

        self._nodes_weight = None
        self._edges_weight = None
        self._element_to_graph = None
        self._pargs = None
        self._expectation_path = []
        self._element_expectation = dict()

    @property
    def element_expectation(self):
        return self._element_expectation

    def get_operator(self, element, qubit_num):
        if self._expectation_calc_method == "statevector":
            op = 1
            for i in range(qubit_num):
                if i in list(element):
                    op = np.kron(Z.to_matrix(), op)
                else:
                    op = np.kron(I.to_matrix(), op)
        else:
            op = Z if 0 in list(element) else I
            for i in range(1, qubit_num):
                if i in list(element):
                    op = Z ^ op
                else:
                    op = I ^ op
        return op

    def get_QAOA_circuit(self, p, graph, node_to_qubit):
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        circ_qreg = QuantumRegister(len(graph), 'q')
        circ = QuantumCircuit(circ_qreg)

        for k in range(p):
            for nd in graph.nodes:
                u = node_to_qubit[nd]
                if k == 0:
                    circ.h(u)
                circ.rz(2 * gamma_list[k] * self._nodes_weight[nd], u)

            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                circ.rzz(2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]], u, v)

            for nd in graph.nodes:
                circ.rx(2 * beta_list[k], node_to_qubit[nd])

        # circ.measure_all()
        return circ

    def get_GHZ_circuit(self, p, graph, node_to_qubit):
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        circ_qreg = QuantumRegister(len(graph), 'q')
        circ = QuantumCircuit(circ_qreg)

        pivot = len(self._origin_graph) // 2

        for k in range(p):
            for nd in graph.nodes:
                if k == 0:
                    circ.h(node_to_qubit[nd])

            for edge in graph.edges:
                    u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                    if u == v:
                        continue
                    circ.rzz(2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]], u, v)  #

            for nd in graph.nodes:
                if nd != pivot:
                    circ.rx(2 * beta_list[k], node_to_qubit[nd])

            # circ.measure_all()
            return circ

    def get_expectation(self, element_graph, p=None):
        if self._is_parallel is False:
            p = self._p if p is None else p
            original_e, graph = element_graph
        else:
            p = self._p if len(element_graph) == 1 else element_graph[1]
            original_e, graph = element_graph[0]

        node_to_qubit = defaultdict(int)
        node_list = list(graph.nodes)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        if self._research == "QAOA":
            circ = self.get_QAOA_circuit(p, graph, node_to_qubit)
        elif self._research == "GHZ":
            pivot = len(self._origin_graph) // 2
            if pivot in graph:
                circ = self.get_GHZ_circuit(p, graph, node_to_qubit)
            else:
                circ = self.get_QAOA_circuit(p, graph, node_to_qubit)
        # circ.draw(output='mpl', interactive=True)
        if isinstance(original_e, int):
            weight = self._nodes_weight[original_e]
            op = self.get_operator([node_to_qubit[original_e]], len(node_list))
        else:
            weight = self._edges_weight[original_e]
            op = self.get_operator([node_to_qubit[original_e[0]], node_to_qubit[original_e[1]]], len(node_list))

        if self._expectation_calc_method == "statevector":
            circ.save_statevector()
            sim = Aer.get_backend('qasm_simulator')  #aer
            result = sim.run(circ, seed_simulator=47, nshots=102400).result()
            out_state = result.get_statevector()
            exp_res = np.matmul(np.matmul(out_state.conj().T, op), out_state).real

        elif self._expectation_calc_method == "sample":
            subc = CircuitStateFn(circ)
            backend = Aer.get_backend('qasm_simulator')
            q_instance = QuantumInstance(backend, shots=1024)
            measurable_expression = StateFn(op, is_measurement=True).compose(subc)
            expectation = PauliExpectation().convert(measurable_expression)
            sampler = CircuitSampler(q_instance).convert(expectation)
            exp_res = sampler.eval().real

        elif self._expectation_calc_method == "statistic":
            circ.measure_all()
            backend = Aer.get_backend('qasm_simulator')
            backend.shots = 1024
            counts = backend.run(circ, seed_simulator=10, nshots=1024 * 100).result().get_counts()
            # counts = backend.run(circ, seed_simulator=43, nshots=1024).result().get_counts()

            def Ising_obj(x, G):
                obj = 0
                for ed in G.edges():
                    if x[node_to_qubit[ed[0]]] == x[node_to_qubit[ed[1]]]:
                        obj -= self._edges_weight[ed]
                    else:
                        obj += self._edges_weight[ed]
                return obj

            avg = 0
            sum_count = 0
            for bitstring, count in counts.items():
                obj = Ising_obj(bitstring, graph)
                avg += obj * count
                sum_count += count
            exp_res = avg / sum_count

        return weight, exp_res

    def expectation_calculation(self, p=None):
        if self._nodes_weight is None or self._edges_weight is None:
            nodes_weight, edges_weight = get_graph_weights(self._origin_graph)
            self._nodes_weight, self._edges_weight = nodes_weight, edges_weight

        self._element_expectation = {}
        if self._is_parallel:
            return self.expectation_calculation_parallel(p)
        else:
            return self.expectation_calculation_serial(p)

    def expectation_calculation_serial(self, p=None):
        cpu_num = cpu_count()
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        res = 0
        # st = time.time()
        for item in self._element_to_graph.items():
            w_i, exp_i = self.get_expectation(item, p)
            if isinstance(item[0], tuple):  #origin format of RQAOA
                self._element_expectation[item[0]] = exp_i
            res += w_i * exp_i  #
        # ed = time.time()
        # print("exp one cost:", ed - st)
        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def expectation_calculation_parallel(self, p=None):
        cpu_num = 1
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        circ_res = []
        args = list(itertools.product(self._element_to_graph.items(), [p]))
        pool = Pool(os.cpu_count())
        circ_res.append(pool.map(self.get_expectation, args))
        # circ_res.append(pool.map(self.get_expectation, list(self._element_to_graph.items()), chunksize=1))

        pool.terminate()  # pool.close()
        pool.join()
        res = 0
        for it in circ_res[0]:
            res += it[0] * it[1]

        # res = sum(circ_res[0])
        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def optimization_visualization(self):
        plt.figure()
        plt.plot(range(1, len(self._expectation_path) + 1), self._expectation_path, "ob-", label="Qiskit")
        plt.ylabel('Expectation value')
        plt.xlabel('Number of iterations')
        plt.legend()
        plt.show()

    # below is functions not used now
    def graph_to_circuit(self, params, graph, p=None, original_e=None):
        """
        transform the graph to circuit according to the computing_framework
        Args:
            graph (nx.Graph): graph to be transformed to circuit
            params (List): Optimal parameters   #np.array
            original_e (Optional[None, int, tuple])
        Return:
            if original_e=None, then the graph is the whole original graph generated by
            generate_weighted_graph(), so just return the circuit transformed by it

            if original_e is a int, then the subgraph is generated by node(idx = original_e
            in whole graph), so return the it's idx mapped by node_to_qubit[], and the circuit

            if original_e is a tuple, then the subgraph is generated by edge(node idx = original_e
            in whole graph), so return the it's idx mapped by node_to_qubit[] as
            tuple(mapped node_id1, mapped node_id2), and the circuit
        """

        # generate node_to_qubit[] map from node id in graph to [0, node_num in graph - 1]
        # so the subgraph would has it's own qubits number
        p = self._p if p is None else p

        node_to_qubit = defaultdict(int)
        node_list = list(graph.nodes)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        self._pargs = params
        if self._research == "QAOA":
            circ = self.get_QAOA_circuit(p, graph, node_to_qubit)
        elif self._research == "GHZ":
            circ = self.get_GHZ_circuit(p, graph, node_to_qubit)

        if original_e is None:
            return original_e, circ
        elif isinstance(original_e, int):
            return node_to_qubit[original_e], circ
        else:
            return (node_to_qubit[original_e[0]], node_to_qubit[original_e[1]]), circ

    def get_result_statevector(self, params, graph):

        params = params.flatten()
        _, circ = self.graph_to_circuit(params, graph)
        circ.save_statevector()
        # circ.draw(output='mpl', interactive=True)

        sim = Aer.get_backend('aer_simulator')
        result = sim.run(circ).result()
        out_state = result.get_statevector()

        return out_state

    def get_result_counts(self, params):  # 直接将最优参数值带入整个大电路中运行qubit将只能到18

        params = params.flatten()
        _, circ = self.graph_to_circuit(params, self._origin_graph)
        # circ.draw(output='mpl', interactive=True)
        circ.measure_all()   # if use sample method, can't use measure

        sim = Aer.get_backend('qasm_simulator')
        transpiled_qc = transpile(circ, sim)
        job_sim = sim.run(transpiled_qc, shots=1024)
        hist = job_sim.result().get_counts(transpiled_qc)
        return hist

    def sampling_visualization(self, counts):
        from qiskit.visualization import plot_histogram
        plot_histogram(counts)
        plt.show()



# This code is part of Qcover.
#
# (C) Copyright BAQIS 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import logging
import numpy as np
import random
import networkx as nx
import matplotlib.pyplot as plt
import time, os
from numbers import Number
import pandas as pd
from numpy.random import choice
from scipy import optimize
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_number_list

logger = logging.getLogger(__name__)


class General01Programming:
    """
    General 01 Programming:
    problems in the form (Max cx, s.t. Ax=b and/or inequality constraints, x binary)
    Any problem in linear constraints and bounded integer variables can be
    converted through a binary expansion into QUBO model
    """

    def __init__(self,
                 element_list: list = None,
                 number_list: list = None,
                 length: int = None,
                 weight_range: tuple = (1, 100),
                 weight: list = None,
                 element_set: list = None,  # coefficient of the constraints
                 signs: list = None,
                 b: np.array = None,
                 P: float = None,
                 slack_1: int = 3,
                 seed: int = None):
        """
        Args:
            element_list (list): a list of elements 
            length (int): length of element_list
            weight (list): list of weight for element in number_list
            element_set (list): a list of coefficient of constraints
            signs (list): a list of constraint equations/inequalities signs
            b (list): a list of constraint equations/inequalities values
            P (int): the penalty value for the penalty terms
            slack_1 (int): slack variable for '<=' inequalty constraints, default value is 3
            (require input: number_list, weight, element_set, signs, b, P, slack_1(optional))

         Returns:
            element_list, length, weight_range, weight, element_set, signs, constraints, b, P

        Example:
            element_list = ['a','b','c','d','e']
            element_list_len = len(element_list)
            element_weight = [6,4,8,5,5]
            coefficients = [[2,2,4,3,2],[1,2,2,1,2],[3,3,2,4,4]]
            signs = ['<=', '=', '>=']
            constants = [7,4,5]
            penalty = 10
            slack_1 = 3
        """
        self._slack_1 = slack_1
        
        for n in [element_list]:
            if isinstance(n, Number) == False:
                number_list = pd.factorize(element_list)[0] 
                
        if number_list is None and element_set is None:
            assert length is not None
            assert len(element_set) is not None
            self._length = length
            self._weight_range = weight_range
            self._weight = weight
            self._element_set = element_set
            self._signs = signs
            self._b = b
            self._P = P
            self._constraints = self.get_constraints()
            self._seed = seed
            self._element_list = random_number_list(n=self._length,
                                                   weight_range=self._weight_range,
                                                   seed=self._seed)  

        else:
            for n in [element_list]:
                if isinstance(n, Number) == False:
                    self._element_list = pd.factorize(element_list)[0] 
            self._length = len(self._element_list)
            self._weight_range = (np.abs(self._element_list).min(), np.abs(self._element_list).max())
            self._weight = weight
            self._element_set = element_set
            self._signs = signs
            self._b = b   
            self._P = P
            self._constraints = self.get_constraints()
            self._seed = seed

        self._qmatrix = None
        self._shift = None

    @property
    def length(self):
        return self._length

    @property
    def weight(self):
        return self._weight

    def get_constraints(self):
        """
        convert all constraints into equalities of the form Ax=b
        and get the A matrix from the constraints

        Args:
            element_set (list) : a list of resource requirement of project
            slack (int): slack variable coefficient, default value is 3
        Returns:
            A (np.array): the matrix of constraint coefficients

        ..math::
            if signs[l] is '=', A_list[l] equals to element_set[l]

            if signs[j] is '<=', the default slack variable slack_1
            upper bound default value is 3, if slack_1 is an even number, it should be converted
            to an odd number by adding 1. slack_1 is then binary expanded into slack_cof1,
            A_list[j] equals to element_set[l] extended by slack_cof1.

            if signs[k] is '>=', the default slack variable slack_2
            upper bound default value is b[k]+1, if slack_2 is an even number, it should be converted
            to an odd number by adding 1. slack_2 is then binary expanded into slack_cof2.
            A_list[k] equals to element_set[k]

            A_list[l], A_list[j], A_list[k] are extended by 0s,
            A_list[k] is extended by slack_cof2 such that A_list elements have the same length.
        """
        A_list = [[] for i in range(len(self._element_set))]

        index_1 = [i for i in range(len(self._signs)) if self._signs[i] == '=']
        l = int(index_1[0])
        A_list[l] = list(self._element_set[l])

        index_2 = [i for i in range(len(self._signs)) if self._signs[i] == '<=']
        j = int(index_2[0])
        y = list([int(i) for i in bin(self._slack_1)[2:]])  # binary expansion for slack variable
        y.reverse()
        if self._slack_1 % 2 == 0:
            y = list([int(i) for i in bin(self._slack_1 + 1)[2:]])
        # the coefficients of new binary variables resulted from the slack variable
        slack_cof1 = np.multiply(np.array(y), np.float_power(2, np.arange(len(y)))).tolist()
        A_list[j] = list(self._element_set[j])
        A_list[j].extend(int(i) for i in slack_cof1)

        index_3 = [i for i in range(len(self._signs)) if self._signs[i] == '>=']
        k = int(index_3[0])
        slack_2 = self._b[k] + 1  # the upper bound of the slack variable
        z = list([int(i) for i in bin(slack_2)[2:]])  # binary expansion for slack variable
        z.reverse()
        if slack_2 % 2 == 0:
            z = list([int(i) for i in bin(slack_2 + 1)[2:]])
        # the coefficients of new binary variables resulted from the slack variable
        slack_cof2 = (-1.0 * np.multiply(np.array(z), np.float_power(2, np.arange(len(z))))).tolist()
        A_list[k] = list(self._element_set[k])

        A_list[l].extend([0.0] * (len(slack_cof1) + len(slack_cof2)))
        A_list[j].extend([0.0] * (len(slack_cof2)))
        A_list[k].extend([0.0] * (len(slack_cof1)))
        A_list[k].extend(int(i) for i in slack_cof2)

        return np.array(A_list)

    def update_args(self, length, weight, A):
        self._length = length
        self._weight = weight

    def get_Qmatrix(self):
        """
        get the Q matrix in QUBO model of number partition problem

        Args:
            numbers (np.array): the number set that to be divided

        Returns:
            q_mat (np.array): the the Q matrix of QUBO model.

        ..math::
            minimise x(T)Qx
            Q[i][j] = P/2 * A(T)A[i][j]
            Q[i][i] = -weight[i] + P * (A(T)A[i][i] - 2A(T)b[i])
        """
        matrix_dimension = self._constraints.shape[1]
        q_mat = np.eye(matrix_dimension, dtype='float64')

        extended_weight = list(self._weight)
        extended_weight.extend([0] * (self._constraints.shape[1] - self._length))

        A_matrix_mul = np.matmul(np.transpose(self._constraints), self._constraints)  # A(T)A
        AT_b = np.matmul(np.transpose(self._constraints), self._b)  # A(T)b

        for i in range(matrix_dimension):
            q_mat[i][i] = -extended_weight[i] + self._P * (A_matrix_mul[i][i] - 2.0 * AT_b[i])
            for j in range(matrix_dimension):
                if i == j:
                    continue
                q_mat[i][j] = self._P * A_matrix_mul[i][j]
        
        shift = self._P * np.matmul(np.transpose(self._b), self._b)

        return q_mat, shift

    def general_01_programming_value(self, x, w):
        """Compute the value of the problem.

        Args:
            x (numpy.ndarray): binary string as numpy array.
            weight (numpy.ndarray): weights of elements

        Returns:
            float: value of the problem.
        """
        if self._qmatrix is None:
            self._qmatrix = self.get_Qmatrix()

        X = np.matmul(x, np.matmul(self._qmatrix, np.transpose(x)))
        return X

    def run(self):
        if self._qmatrix is None:
            self._qmatrix, self._shift = self.get_Qmatrix()

        qubo_mat = self._qmatrix
        ising_mat = get_ising_matrix(qubo_mat)
        gp_graph = get_weights_graph(ising_mat)
        return gp_graph, self._shift
    


from .optimizer import Optimizer
from .COBYLA import COBYLA
from .SLSQP import SLSQP
from .L_BFGS_B import L_BFGS_B
from .Gradient_Descent import GradientDescent
from .Interp import Interp
from .Fourier import Fourier
from .SPSA import SPSA
from .SHGO import SHGO

__all__ = ['Optimizer',
           'COBYLA',
           'SLSQP',
           'L_BFGS_B',
           'GradientDescent',
           'Interp',
           'Fourier',
           'SPSA',
           'SHGO'
           ]